#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <afblib/shared_domain.h>
#include <afblib/shared_rts.h>
#include "master.h"
#include "parameters.h"
#include "rgbcolor.h"
#include "worker.h"

#define RGB(r, g, b) \
   {.red = (double)r / 255, .green = (double)g / 255, .blue = (double)b / 255}

RGBColor sunset[] = {
   RGB(25, 7, 26),
   RGB(9, 1, 47),
   RGB(4, 4, 73),
   RGB(0, 7, 100),
   RGB(12, 44, 138),
   RGB(24, 82, 177),
   RGB(57, 125, 209),
   RGB(134, 181, 229),
   RGB(211, 236, 248),
   RGB(241, 233, 191),
   RGB(248, 201, 95),
   RGB(255, 170, 0),
   RGB(204, 128, 0),
   RGB(153, 87, 0),
   RGB(106, 52, 3),
   RGB(66, 30, 15),
};

static Parameters p = {
   .center_x = -0.7,
   .center_y = 0,
   .horizontal_diameter = 3.0769,
   .width = 256,
   .height = 192,
   .maxiterations = 1000,
   .color_shift = 0,
   .color_progression = 200,
   .colors = sunset,
   .nof_colors = sizeof(sunset)/sizeof(sunset[0]),
};

#define MEMBER(field) ((void*)&p.field)

bool read_double(char* arg, void* p) {
   return sscanf(arg, "%lf", (double*) p) == 1;
}

bool read_uint(char* arg, void* p) {
   return sscanf(arg, "%u", (unsigned int*) p) == 1;
}

static struct Options {
   char name;
   const char* description;
   bool (*input)(char* arg, void* p);
   void* addr;
} options[] = {
   {'x', "center (real part)", read_double, MEMBER(center_x)},
   {'y', "center (imaginary part)", read_double, MEMBER(center_y)},
   {'d', "horizontal diameter", read_double, MEMBER(horizontal_diameter)},
   {'w', "width in pixels", read_uint, MEMBER(width)},
   {'h', "height in pixels", read_uint, MEMBER(height)},
   {'m', "maximal # of iterations", read_uint, MEMBER(maxiterations)},
   {'s', "color shift", read_double, MEMBER(color_shift)},
   {'P', "color progression",
      read_double, MEMBER(color_progression)},
};
static const unsigned int nofoptions = sizeof(options)/sizeof(options[0]);

static char* cmdname = 0;

void usage() {
   fprintf(stderr, "Usage: %s", cmdname);
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, " [-%c value]", options[i].name);
   }
   fprintf(stderr, " outfile\n");
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, "   -%c value: %s\n",
	 options[i].name, options[i].description);
   }
   exit(1);
}

void noop(int signo) {
   /* does nothing */
}

int main(int argc, char** argv) {
   cmdname = *argv++; --argc;
   struct shared_domain* sd = shared_rts_init();
   if (!sd) {
      fprintf(stderr, "%s: was not invoked by smrun\n", cmdname);
      exit(1);
   }
   unsigned int rank = sd_get_rank(sd);
   unsigned int nofprocesses = sd_get_nofprocesses(sd);
   if (nofprocesses == 1) {
      fprintf(stderr, "%s: more processes are required\n", cmdname);
      exit(1);
   }

   /* process arguments */
   char* outfile = 0;
   if (rank == 0) {
      while (argc > 0 && argv[0][0] == '-') {
	 char option = argv[0][1];
	 unsigned int i = 0;
	 while (i < nofoptions && option != options[i].name) ++i;
	 if (i == nofoptions) usage();
	 char* val;
	 if (argv[0][2]) {
	    val = &argv[0][2];
	 } else {
	    --argc; ++argv;
	    if (argc == 0) usage();
	    val = *argv;
	 }
	 if (!options[i].input(val, options[i].addr)) usage();
	 --argc; ++argv;
      }
      if (argc != 1) usage();
      outfile = *argv++; --argc;

      /* compute derived parameters */
      p.resolution = p.width / p.horizontal_diameter;
      p.xmin = p.center_x - p.horizontal_diameter / 2;
      p.xmax = p.center_x + p.horizontal_diameter / 2;
      p.vertical_diameter = p.height / p.resolution;
      p.ymin = p.center_y - p.vertical_diameter / 2;
      p.ymax = p.center_y + p.vertical_diameter / 2;

      /* compute required space */
      size_t pixelbuf_size = sizeof(guchar) * 3 * p.width * p.height;
      size_t required = sizeof(Parameters) + pixelbuf_size;
      size_t extra_space_size = sd_get_extra_space_size(sd);
      if (extra_space_size < required) {
	 fprintf(stderr, "%s: "
	    "just %zu bytes extra space were provided "
	    "but %zu bytes are required\n",
	    cmdname, extra_space_size, required);
	 exit(1);
      }
   }
   void* extra = sd_get_extra_space(sd);
   Parameters* shared_params = (Parameters*) extra;
   guchar* pixelbuf = (guchar*) ((char*) extra + sizeof(Parameters));
   if (rank == 0) {
      *shared_params = p;
   }
   sd_barrier(sd);
   Parameters params = *shared_params;
   params.colors = sunset;

   if (rank == 0) {
      if (!master(sd, &params)) {
	 perror("master"); exit(1);
      }
   } else {
      if (!worker(sd, &params, pixelbuf)) {
	 perror("worker"); exit(1);
      }
   }

   sd_barrier(sd);
   if (rank == 0) {
      /* generate output */
      GdkPixbuf* gdk_pixbuf = gdk_pixbuf_new_from_data(
	 /* data = */ pixelbuf,
	 /* colorspace = */ GDK_COLORSPACE_RGB,
	 /* has_alpha = */ FALSE,
	 /* bits_per_sample = */ 8,
	 /* width = */ p.width,
	 /* height = */ p.height,
	 /* rowstride = */ sizeof(guchar) * 3 * p.width,
	 /* destroy_fn = */ NULL,
	 /* destroy_fn_data = */ NULL);
      gdk_pixbuf_save(gdk_pixbuf, outfile, "jpeg", (GError **)0,
	 "quality", "100", (void*)0);
   }
}
